# aiomatter/bot.py

import asyncio
from logging import Logger
from typing import Any, Callable, Coroutine, Dict, List

from .driver import MattermostDriver
from .events import EventType
from .plugin import Plugin
from .settings import Settings
from .aiologger import setup_logger


EventHandler = Callable[[Dict[str, Any]], Coroutine[Any, Any, None]]


class Bot:
    def __init__(
        self,
        plugins: List[Plugin],
        settings: Settings,
        logger: Logger | None = None,
    ):
        """Инициализация бота с подключенными плагинами и настройками."""
        self.plugins = plugins
        self.settings = settings
        self.logger = logger or setup_logger()
        self.driver = MattermostDriver(
            settings.full_api_url, settings.BOT_TOKEN, logger=self.logger
        )
        self.handlers: Dict[EventType, List[EventHandler]] = (
            self._register_plugins()
        )

    def _register_plugins(self) -> Dict[EventType, List[EventHandler]]:
        """Регистрирует плагины и передает им драйвер."""
        handlers: Dict[EventType, List[EventHandler]] = {}
        for plugin in self.plugins:
            plugin.setup_plugin(self.driver, self.logger)
            for event_type, funcs in plugin.get_handlers().items():
                handlers.setdefault(event_type, []).extend(funcs)
        return handlers

    async def _initialize(self) -> None:
        await self.driver.initialize()

    async def handle_event(self, event: Dict[str, Any]):
        """Обрабатывает событие и вызывает соответствующие хэндлеры."""
        try:
            event_type = event.get('event')
            matching_handlers = self.handlers.get(event_type, [])

            if EventType.ANY in self.handlers:
                matching_handlers.extend(self.handlers[EventType.ANY])

            for handler in matching_handlers:
                await handler(event)
        except Exception as e:
            self.logger.exception(f"Ошибка при обработке события:\n{e}")

    async def _async_run(self):
        """Асинхронный метод запуска бота."""
        await self._initialize()

        ws_url = f"{self.settings.full_api_url}/websocket"
        self.logger.info(f"Бот подключается к WebSocket: {ws_url}")

        while True:
            try:
                await self.driver.connect_websocket(ws_url, self.handle_event)
            except Exception as e:
                self.logger.error(
                    f"Ошибка WebSocket: {e}. Переподключение через 5 секунд..."
                )
                await asyncio.sleep(5)

    def run(self):
        """Синхронный метод запуска бота."""
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        try:
            loop.run_until_complete(self._async_run())
        finally:
            loop.close()

    async def _async_stop(self):
        """Асинхронная остановка бота."""
        await self.driver.close()

    def stop(self):
        """Синхронный метод для остановки бота."""
        asyncio.run(self._async_stop())



import json
from aiomatter.schemas.post_events import PostEvent, EditEvent, DeleteEvent
from aiomatter.events import EventType

EVENT_MAPPING = {
    EventType.POSTED: PostEvent,
    EventType.POST_EDITED: EditEvent,
    EventType.POST_DELETED: DeleteEvent,
    EventType.ANY: lambda event: event,
}


def listen(event_type: EventType, ignore_bots: bool = True):
    """Декоратор для регистрации обработчиков событий."""

    def decorator(func):
        async def wrapper(plugin, event):
            post_data = json.loads(event['data'].get('post', '{}'))
            sender = post_data.get('sender_name', None)
            props = post_data.get('props', {})
            from_bot = props.get('from_bot', 'false')

            if (from_bot == 'true' or sender == 'System') and ignore_bots:
                return

            event_class = EVENT_MAPPING.get(event_type)
            if not event_class:
                raise ValueError(f"Неизвестный тип события: {event_type}")

            if event_type == EventType.ANY:
                typed_event = event
            else:
                typed_event = event_class(**event)

            await func(plugin, typed_event)

        wrapper.event_type = event_type
        return wrapper

    return decorator



# aiomatter/driver.py

import json
from logging import Logger

import aiohttp
import websockets

from aiomatter.schemas.post_events import Post
from aiomatter.schemas.user import MattermostUser


class AsyncDriver:
    def __init__(self, base_url, token, logger: Logger):
        self.base_url = base_url
        self.token = token
        self.logger = logger
        self.session = None

    async def initialize(self):
        """Асинхронная инициализация сессии."""
        self.session = aiohttp.ClientSession(
            headers={"Authorization": f"Bearer {self.token}"}
        )

    async def send_request(self, endpoint, method='GET', data=None):
        """Отправляет запрос к REST API."""
        url = f"{self.base_url}{endpoint}"
        async with self.session.request(method, url, json=data) as response:
            response.raise_for_status()
            return await response.json()

    async def connect_websocket(self, ws_url: str, on_message):
        """Устанавливает WebSocket-соединение и слушает события."""
        headers = {'Authorization': f'Bearer {self.token}'}
        try:
            async with websockets.connect(ws_url, extra_headers=headers) as ws:
                self.logger.info("Успешно подключено к WebSocket.")
                async for message in ws:
                    try:
                        self.logger.debug(f"Получено сообщение: {message}")
                        await on_message(json.loads(message))
                    except Exception as e:
                        self.logger.error(
                            f"Ошибка при обработке сообщения: {e}"
                        )
        except websockets.exceptions.ConnectionClosed as e:
            self.logger.error(f"Соединение закрыто: {e}")
            raise e
        except Exception as e:
            self.logger.error(f"Ошибка WebSocket: {e}")
            raise e

    async def close(self):
        """Закрывает сессию для освобождения ресурсов."""
        await self.session.close()


class MattermostDriver(AsyncDriver):
    """Драйвер взаимодействия с Mattermost API на основе WebSocket."""

    async def send_message(self, channel_id, message, root_id=None):
        """Отправляет сообщение в указанный канал."""
        data = {"channel_id": channel_id, "message": message}

        if root_id is not None:
            data["root_id"] = root_id

        response = await self.send_request("/posts", method='POST', data=data)

        return Post(**response)

    async def edit_message(self, post_id, new_message):
        """Редактирует существующее сообщение по его ID."""
        data = {"message": new_message}
        response = await self.send_request(
            f"/posts/{post_id}/patch", method='PUT', data=data
        )

        return Post(**response)

    async def delete_message(self, post_id):
        """Удаляет сообщение по его ID."""
        await self.send_request(f"/posts/{post_id}", method='DELETE')

    async def _create_dm_channel(self, user_id):
        """Создает канал для личного сообщения с пользователем."""
        bot_info = await self.get_bot_info()
        data = [bot_info.id, user_id]
        response = await self.send_request(
            "/channels/direct", method='POST', data=data
        )
        return response["id"]

    async def direct_message(self, user_id, message):
        """Отправляет личное сообщение пользователю."""
        channel_id = await self._create_dm_channel(user_id)
        return await self.send_message(channel_id, message)

    async def get_user_by_name(self, username: str) -> MattermostUser:
        """Получает информацию о пользователе по его имени."""
        username = username.replace('@', '')
        response = await self.send_request(
            f"/users/username/{username}", method='GET'
        )
        return MattermostUser(**response)

    async def get_user_by_id(self, user_id) -> MattermostUser:
        """Получает информацию о пользователе по его ID."""
        response = await self.send_request(f"/users/{user_id}", method='GET')
        return MattermostUser(**response)

    async def get_bot_info(self) -> MattermostUser:
        """Возвращает информацию о боте."""
        response = await self.send_request(
            '/users/me',
            method='GET',
        )
        return MattermostUser(**response)



# aiomatter/plugin.py

from logging import Logger
from aiomatter.driver import MattermostDriver


class Plugin:
    def __init__(self):
        self.driver: MattermostDriver = None
        self.logger: Logger = None
        self.handlers = self._load_handlers()

    def _load_handlers(self):
        """Собирает все методы-обработчики с декораторами."""
        handlers = {}
        for attr_name in dir(self):
            attr = getattr(self, attr_name)
            if callable(attr) and hasattr(attr, 'event_type'):
                event_type = attr.event_type
                handlers.setdefault(event_type, []).append(attr)
        return handlers

    def setup_plugin(self, driver, logger):
        """Устанавливает драйвер для плагина."""
        self.driver = driver
        self.logger = logger

    def get_handlers(self):
        """Возвращает все зарегистрированные обработчики."""
        return self.handlers



import logging
import os

LOG_FILE_PATH = os.path.join(os.getcwd(), "aiomatter.log")


def setup_logger(
    name: str = "Aiomatter", log_file: str = LOG_FILE_PATH
) -> logging.Logger:
    """Создает и настраивает логгер."""
    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)

    log_format = logging.Formatter(
        "[%(asctime)s] %(name)s: %(message)s", datefmt="%d.%m.%Y %H:%M:%S"
    )

    file_handler = logging.FileHandler(log_file, encoding="utf-8")
    file_handler.setFormatter(log_format)

    console_handler = logging.StreamHandler()
    console_handler.setFormatter(log_format)

    logger.addHandler(file_handler)
    logger.addHandler(console_handler)

    logger.propagate = False

    return logger



from typing import Union

from pydantic import Field
from pydantic.dataclasses import dataclass


@dataclass
class UserProps:
    custom_status: Union[str, None] = Field(default=None, alias='customStatus')


@dataclass
class NotifyProps:
    auto_responder_active: Union[str, None] = Field(
        default=None, alias='auto_responder_active'
    )
    auto_responder_message: Union[str, None] = Field(
        default=None, alias='auto_responder_message'
    )
    calls_desktop_sound: Union[str, None] = Field(
        default=None, alias='calls_desktop_sound'
    )
    calls_notification_sound: Union[str, None] = Field(
        default=None, alias='calls_notification_sound'
    )
    channel: Union[str, None] = Field(default=None)
    comments: Union[str, None] = Field(default=None)
    desktop: Union[str, None] = Field(default=None)
    desktop_notification_sound: Union[str, None] = Field(
        default=None, alias='desktop_notification_sound'
    )
    desktop_sound: Union[str, None] = Field(default=None)
    desktop_threads: Union[str, None] = Field(default=None)
    email: Union[str, None] = Field(default=None)
    email_threads: Union[str, None] = Field(default=None)
    first_name: Union[str, None] = Field(default=None, alias='first_name')
    highlight_keys: Union[str, None] = Field(
        default=None, alias='highlight_keys'
    )
    mention_keys: Union[str, None] = Field(default=None, alias='mention_keys')
    push: Union[str, None] = Field(default=None)
    push_status: Union[str, None] = Field(default=None, alias='push_status')
    push_threads: Union[str, None] = Field(default=None)


@dataclass
class Timezone:
    automatic_timezone: Union[str, None] = Field(
        default=None, alias='automaticTimezone'
    )
    manual_timezone: Union[str, None] = Field(
        default=None, alias='manualTimezone'
    )
    use_automatic_timezone: Union[str, None] = Field(
        default=None, alias='useAutomaticTimezone'
    )


@dataclass
class MattermostUser:
    id: str
    create_at: int
    update_at: int
    delete_at: int
    username: str
    auth_data: str
    auth_service: str
    email: str
    nickname: str
    first_name: str
    last_name: str
    roles: str  # Поля без значения по умолчанию идут выше
    locale: str
    disable_welcome_email: bool
    position: Union[str, None] = (
        None  # Поля с значением по умолчанию идут ниже
    )
    props: UserProps = Field(default_factory=UserProps)
    notify_props: NotifyProps = Field(default_factory=NotifyProps)
    last_picture_update: Union[int, None] = None
    timezone: Timezone = Field(default_factory=Timezone)

    class Config:
        populate_by_name = True

    def get_full_name(self) -> str:
        return f"{self.last_name} {self.first_name}"

    def is_system_admin(self) -> bool:
        return 'system_admin' in self.roles

    def is_channel_admin(self) -> bool:
        return 'channel_admin' in self.roles



# aiomatter/events.py

from enum import StrEnum


class EventType(StrEnum):
    # ADDED_TO_TEAM = "added_to_team"
    # AUTHENTICATION_CHALLENGE = "authentication_challenge"
    # CHANNEL_CONVERTED = "channel_converted"
    # CHANNEL_CREATED = "channel_created"
    # CHANNEL_DELETED = "channel_deleted"
    # CHANNEL_MEMBER_UPDATED = "channel_member_updated"
    # CHANNEL_UPDATED = "channel_updated"
    # CHANNEL_VIEWED = "channel_viewed"
    # CONFIG_CHANGED = "config_changed"
    # DELETE_TEAM = "delete_team"
    # DIRECT_ADDED = "direct_added"
    # EMOJI_ADDED = "emoji_added"
    # EPHEMERAL_MESSAGE = "ephemeral_message"
    # GROUP_ADDED = "group_added"
    # HELLO = "hello"
    # LEAVE_TEAM = "leave_team"
    # LICENSE_CHANGED = "license_changed"
    # MEMBERROLE_UPDATED = "memberrole_updated"
    # NEW_USER = "new_user"
    # PLUGIN_DISABLED = "plugin_disabled"
    # PLUGIN_ENABLED = "plugin_enabled"
    # PLUGIN_STATUSES_CHANGED = "plugin_statuses_changed"
    POST_DELETED = "post_deleted"
    POST_EDITED = "post_edited"
    # POST_UNREAD = "post_unread"
    POSTED = "posted"
    # PREFERENCE_CHANGED = "preference_changed"
    # PREFERENCES_CHANGED = "preferences_changed"
    # PREFERENCES_DELETED = "preferences_deleted"
    # REACTION_ADDED = "reaction_added"
    # REACTION_REMOVED = "reaction_removed"
    # RESPONSE = "response"
    # ROLE_UPDATED = "role_updated"
    # STATUS_CHANGE = "status_change"
    # TYPING = "typing"
    # UPDATE_TEAM = "update_team"
    # USER_ADDED = "user_added"
    # USER_REMOVED = "user_removed"
    ANY = "*"



# aimatter/settings.py


class Settings:
    def __init__(
        self, bot_token: str, base_url: str, api_path: str = "/api/v4"
    ):
        """Инициализация параметров конфигурации с автоматической заменой схемы."""
        self.BOT_TOKEN = bot_token
        self.BASE_URL = self._transform_base_url(base_url.rstrip('/'))
        self.API_PATH = api_path

    @staticmethod
    def _transform_base_url(base_url: str) -> str:
        """Заменяет http/https на ws/wss в URL."""
        if base_url.startswith("https://"):
            return base_url.replace("https://", "wss://", 1)
        elif base_url.startswith("http://"):
            return base_url.replace("http://", "ws://", 1)
        else:
            raise ValueError(
                "Некорректная схема URL. Ожидается http:// или https://"
            )

    @property
    def full_api_url(self) -> str:
        """Возвращает полный URL для API."""
        return f"{self.BASE_URL}{self.API_PATH}"



import json
from typing import Any, Dict, Union

from pydantic import Field, field_validator
from pydantic.dataclasses import dataclass

from aiomatter.events import EventType


@dataclass
class Post:
    id: str
    create_at: int
    update_at: int
    edit_at: int
    delete_at: int
    is_pinned: bool
    user_id: str
    channel_id: str
    root_id: str
    original_id: str
    message: str
    type: str
    hashtags: str
    pending_post_id: str
    reply_count: int
    last_reply_at: int
    remote_id: str | None = None
    props: Dict[str, Any] = Field(default_factory=dict)
    participants: Union[None, Any] = None
    metadata: Dict[str, Any] = Field(default_factory=dict)


@dataclass
class BaseEventData:
    post: Union[Post, str]

    @field_validator("post", mode="before")
    @classmethod
    def parse_post(cls, value: Union[str, Dict[str, Any]]) -> Post:
        """Если 'post' передан как строка, парсим его в объект."""
        if isinstance(value, str):
            try:
                post_data = json.loads(value)
                return Post(**post_data)
            except json.JSONDecodeError:
                raise ValueError("Invalid JSON in 'post' field")
        elif isinstance(value, dict):
            return Post(**value)
        else:
            raise ValueError(f"Invalid type for 'post': {type(value)}")


@dataclass
class DeleteEventData(BaseEventData):
    delete_by: str


@dataclass
class PostEventData(BaseEventData):
    channel_display_name: str
    channel_name: str
    channel_type: str
    sender_name: str
    set_online: bool
    team_id: str


@dataclass
class PostEvent:
    event: EventType
    seq: int
    data: PostEventData
    broadcast: Dict[str, Any] = Field(default_factory=dict)


@dataclass
class EditEvent:
    event: EventType
    seq: int
    data: BaseEventData
    broadcast: Dict[str, Any] = Field(default_factory=dict)


@dataclass
class DeleteEvent:
    event: EventType
    seq: int
    data: DeleteEventData
    broadcast: Dict[str, Any] = Field(default_factory=dict)
